'use strict';

/**
 * @ngdoc function
 * @name rpsControllers.controller:AddEntryCtrl
 * @description
 * # AddEntryCtrl
 * Controlador utilizado para la creación de una nueva entrada.
 */
rpsControllers.controller('AddEntryCtrl', [
    '$scope',
    'EntryService',
    'ModalService',
    '$ionicLoading',
    '$state',
    '$stateParams',
    '$ionicViewService',
    '$jrCrop',

    function($scope, EntryService, ModalService, $ionicLoading, $state, $stateParams, $ionicViewService, $jrCrop) {

        if ($stateParams.entryId) {
            $ionicLoading.show({
                templateUrl: 'templates/loading.html'
            });

            // LLamamos al método getEntryById del servicio para obtener la información de la entrada
            // que se quiere visualizar
            EntryService.getEntryById($stateParams.entryId)
                .then(function (entry) {
                    $scope.entryData = entry;
                    $ionicLoading.hide();
                })
                .fail(function (error) {
                    $scope.modalInfo = {
                        title: 'Ups!',
                        message: error
                    };
                    ModalService.showAlert($scope);
                    $ionicLoading.hide();
                });
        }

        /**
         * Guarda un borrador de la entrada
         * @param entryData {Object} Contiene los datos de la entrada
         */
        $scope.save = function (entryData) {
            save(entryData, false, false);
        };

        /**
         * Guarda un borrador de la entrada y accede a la creación de pasos
         * @param entryData {Object} Contiene los datos de la entrada
         */
        $scope.newStep = function (entryData) {
            save(entryData, true, false);
        };

        /**
         * Guarda y publica una entrada
         * @param entryData {Object} Contiene los datos de la entrada
         */
        $scope.publish = function (entryData) {
            save(entryData, false, true);
        };


        /**
         * Borra un paso existente de la entrada
         * @param step {Object} Contiene los datos del paso
         */
        $scope.removeStep = function (step) {
            if (!step) {
                $scope.modalInfo = {
                    title: 'Ups!',
                    message: 'Hay algún problema para borrar el paso seleccionado. Vuelve al inicio y prueba otra vez.'
                };
                ModalService.showAlert($scope);
            } else {
                $ionicLoading.show({templateUrl: 'templates/deleting.html'});
                EntryService.deleteStep(step.id)
                    .then(function () {
                        $ionicLoading.hide();
                        var numStep = $scope.entryData.steps.indexOf(step);
                        $scope.entryData.steps.splice(numStep, 1);
                        window.plugins.toast.showShortBottom('Paso borrado');
                    })
                    .fail(function (error) {
                        errorFunction(error);
                    });
            }
        };

        /**
         * Hace una llamada al recurso que gestiona la cámara del dispositivo
         * @param type {Number} Indica si la imagen se tomará de la cámara o de la galería
         */
        $scope.takePic = function(type) {
            var options =   {
                saveToPhotoAlbum: true,
                quality: 50,
                destinationType: 1, // 0: DATA_URI, 1: FILE_URI, 2: NATIVE_URI
                sourceType: type,   // 0:Photo Library, 1=Camera, 2=Saved Photo Album
                encodingType: 0     // 0=JPG 1=PNG
            };
            navigator.camera.getPicture(onSuccess, onFail, options);
        };

        /**
         * @private
         * Método privado que se llama desde el método 'takePic' en caso de que se haya tomado
         * la imagen correctamente
         * @param file {Object} Contiene la imagen que se ha tomado del dispositivo
         */
        function onSuccess(file) {
            $jrCrop.crop({
                url: file,
                width: 300,
                height: 240,
                cancelText: 'Cancelar',
                chooseText: 'Recortar',
            }).then(function(canvas) {
                var quality = 0.4;
                $scope.picData = canvas.toDataURL('image/jpeg', quality);
            }, function() {
                // User canceled or couldn't load image.
            });
        }

        /**
         * @private
         * Método privado que se llama desde el método 'takePic' en caso de que se haya producido
         * un error al tomar la imagen del dispositivo
         * la imagen correctamente
         * @param e {Object} Contiene el error que se ha producido
         */
        function onFail(e) {
            console.log(e);
            //errorFunction('Se ha producido un error al obtener la imagen.');
        }

        /**
         * @private
         * Guarda un borrador de la entrada con la opción de acceder a un nuevo paso o publicar la entrada
         * @param entryData {Object} Contiene los datos de la entrada
         * @param newStep {Boolean} Indica si se quiere crear un paso para la entrada
         * @param publish {Boolean} Indica si se quiere publicar la entrada
         */
        function save (entryData, newStep, publish) {
            var picture, entryId;

            // Comprobamos que los datos requeridos para realizar las acciones existen
            if (!entryData || !($scope.picData || entryData.imageThumb) || !entryData.title || !entryData.description) {
                $scope.modalInfo = {
                    title: 'Ups!',
                    message: 'Necesitas añadir la imagen principal, el título y la descripción.'
                };
                ModalService.showAlert($scope);
                return;
            } else {
                if (publish && (!entryData.steps || entryData.steps.length === 0 )) {
                    $scope.modalInfo = {
                        title: 'Ups!',
                        message: 'Necesitas añadir la imagen principal, el título, la descripción y, al menos, un paso para poder publicar la entrada.'
                    };
                    ModalService.showAlert($scope);
                    return;
                }
            }

            // Muestra un mensaje mientras realiza la acción correspondiente
            if (publish) {
                $ionicLoading.show({templateUrl: 'templates/publishing.html'});
            } else {
                $ionicLoading.show({templateUrl: 'templates/saving.html'});
            }

            // Si se ha tomado una nueva imagen se envía para sustotuir a la anterior.
            picture = $scope.picData || undefined;

            // Si existe entryData es porque se esta editando una entrada previamente creada
            if ($scope.entryData) {
                entryId = $scope.entryData.id;
            }

            // Guardamos la entrada
            EntryService.addOrEdit(entryData, picture, entryId)
                .then(function (entry) {
                    $ionicLoading.hide();
                    // Si se quiere crear un nuevo paso, vamos a la vista para añadirlo
                    if (newStep) {
                        $state.go('app.addStep', {entryId: entry.id});
                        window.plugins.toast.showShortBottom('Guardado borrador.');
                    } else if (publish) {
                        EntryService.publish(entryData)
                            .then(function () {
                                $ionicLoading.hide();
                                $ionicViewService.nextViewOptions({
                                    disableBack: true
                                });
                                $state.go('app.wall');
                                window.plugins.toast.showShortBottom('Entrada publicada.');
                            })
                            .fail(function (error) {
                                errorFunction(error);
                            });
                    } else {
                        window.plugins.toast.showShortBottom('Guardado borrador');
                    }

                })
                .fail(function (error) {
                    errorFunction(error);
                });
        }

        /**
         * @private
         * Método que se ejecuta en el caso que se produzca un error para mostrarlo en una
         * ventada modal
         * @param error {error} Contiene el error que se ha producido
         */
        function errorFunction (error) {
            $scope.modalInfo = {
                title: 'Ups!',
                message: error
            };
            ModalService.showAlert($scope);
            $ionicLoading.hide();
        }
    }
]);
