'use strict';

/**
* @ngdoc function
* @name rpsApp.service:EntryService
* @description
* # EntryService
* Servicio para manejar los datos de las entradas.
*/
rpsApp.factory('EntryService', ['UtilsService', 'ErrorsService',

    function(UtilsService, ErrorsService) {

        var service = {},
            Entry = Parse.Object.extend('Entry'),
            Comment = Parse.Object.extend('Comment'),
            Step = Parse.Object.extend('Step'),
            currentEntry;

        /**
         * @private
         * Obtiene los datos de una entrada de Parse.
         * @param entry {Object} Entrada en formato Parse.
         * @return {Object} Objeto con los datos de la entrada.
         */
        function getEntryData (entry) {
            return {
                id: entry.id,
                title: entry.get('title'),
                description: entry.get('description'),
                temporalTags: entry.get('temporalTags'),
                imageThumb: entry.get('imageThumb').url()
            };
        }

        /**
        * @private
        * Obtiene los datos de un comentario de Parse
        * @param entry {Object} Comentario en formato Parse.
        * @return {Object} Objeto con los datos del comentario.
        */
        function getCommentData(comment) {
            return {
                createdBy: comment.get('createdBy').get('username'),
                createdByMD5Email: comment.get('createdBy').get('md5Email'),
                createdAt: comment.createdAt,
                content: comment.get('content')
            };
        }

        /**
         * Añade o edita una entrada en Parse
         * @param entryData {Object} Datos de la entrada
         * @param picData Imagen de la entrada codificada en base64.
         * @return {Object} Los datos de la entrada guardada.
         */
        service.addOrEdit = function (entryData, picData, entryId) {

            var entry = new Entry(),
                user = Parse.User.current(),
                picId = UtilsService.makeId(15),
                base64pic = picData,
                parseFile, title, description, temporalTags;

            if (entryId) {
                // Edicion
                var query = new Parse.Query(Entry);
                return query.get(entryId)
                    .then(function (entry) {
                        entry.set('title', entryData.title);
                        entry.set('description', entryData.description);
                        entry.set('temporalTags', entryData.temporalTags);
                        if (picData) {
                            // Si se ha hecho una nueva imagen se guarda en Parse previamente
                            parseFile = new Parse.File(picId + '.jpg', {base64: base64pic});
                            return parseFile.save()
                                .then(function(parseFile) {
                                    // Si la imagen se ha subido correctamente a Parse se guarda la entrada
                                    entry.set('image', parseFile);
                                    return entry.save()
                                        .then( function(entry) {
                                            return getEntryData (entry);
                                        })
                                        .fail(function(error) {
                                            // En caso de que haya habido algún error al guardar o leer la imagen de Parse, se devuelve el error
                                            return ErrorsService.getErrorMessage(error);
                                        });
                                })
                                .fail(function(error) {
                                    // En caso de que haya habido algún error al guardar o leer la imagen de Parse, se devuelve el error
                                    return ErrorsService.getErrorMessage(error);
                                });
                        } else {
                            // No es necesario guardar otra vez la imagen, puesto que no ha cambiado
                            return entry.save()
                                .then( function(entry) {
                                    return getEntryData (entry);
                                })
                                .fail(function(error) {
                                    // En caso de que haya habido algún error al guardar o leer la imagen de Parse, se devuelve el error
                                    return ErrorsService.getErrorMessage(error);
                                });
                        }
                    })
                    .fail(function (error) {
                        return ErrorsService.getErrorMessage(error);
                    });

            } else {
                // Añadir nueva
                if (entryData) {
                    title = entryData.title || '';
                    description = entryData.description || '';
                    temporalTags = entryData.temporalTags || '';
                }

                entry.set('title', title);
                entry.set('description', description);
                entry.set('createdBy', user);
                entry.set('temporalTags', temporalTags); // tags temporales: los tags finales se crean desde RPSCloud
                entry.set('publish', false); // Nunca se crea una entrada publicada porque no tiene pasos antes de crearla
                parseFile = new Parse.File(picId + '.jpg', {base64: base64pic});
                return parseFile.save()
                    .then(function(savedFile) {
                        // Si la imagen se ha subido correctamente a Parse se guarda la entrada
                        entry.set('image', savedFile);
                        return entry.save()
                            .then( function(entry) {
                                return getEntryData (entry);
                            })
                            .fail(function(error) {
                                // En caso de que haya habido algún error al guardar o leer la imagen de Parse, se devuelve el error
                                return ErrorsService.getErrorMessage(error);
                            });
                    })
                    .fail(function(error) {
                        // En caso de que haya habido algún error al guardar o leer la imagen de Parse, se devuelve el error
                        return ErrorsService.getErrorMessage(error);
                    });

            }
        };

        /**
         * Publica una entrada para que sea visible a todos los usuarios
         * @param entryId {String} identificador de la entrada
         * @return {Parse.Promise} Promesa que se resuelve cuando el objeto se salva.
         */
        service.publish = function (entry) {

            var query = new Parse.Query(Entry);

            return query.get(entry.id)
                .then(function (entry) {
                    entry.set('publish', true);

                    return entry.save();
                })
                .fail(function (error) {
                    return ErrorsService.getErrorMessage(error);
                });
        };


        /**
         * Devuelve la lista de entradas.
         * @param {Number} page Página que queremos obtener.
         * @return {Object} Lista de entradas devueltos por el servicio.
         */
        service.list = function (page) {
            var getPage = page || 0;

            return Parse.Cloud.run('getAllEntries', {page: getPage})
                .then(function(results) {
                    return results;
                })
                .fail(function (error) {
                    return ErrorsService.getErrorMessage(error);
                });
        };

        /**
        * Obtiene una entrada en base a su Id.
        * @param {String} entryId Identificador de la entrada que se quiere obtener.
        * @return {Object} Objeto que representa la entrada.
        */
        service.getEntryById = function (entryId) {
            return Parse.Cloud.run('getEntryById', {entryId: entryId})
                .then(function(results) {
                    if (results.result === 'ok') {
                        currentEntry = results.entry;
                        return results.entry;
                    }
                })
                .fail(function (error) {
                    return ErrorsService.getErrorMessage(error);
                });
        };

        /**
         * Obtiene las entradas de un usuario.
         * @param page {Number} pagina que queremos obtener.
         * @param user {String} user Identificador del usuario del que queremos obtener las entradas.
         * @return {Object} Lista de entradas del usuario.
         */
        service.getUserEntries = function (page, user) {
            var getPage = page || 0;

            return Parse.Cloud.run('getAllEntries', {page: getPage, userId: user})
                .then(function(results) {
                    return results;
                })
                .fail(function (error) {
                    return ErrorsService.getErrorMessage(error);
                });
        },

        /**
         * Añade un comentario a la entrada actual.
         * @param entryId {String} Identificador de la entrada a la que se desea añadir el comentario.
         * @param content {String} Texto del comentario.
         * @return {Object} Objeto con los datos del comentario.
         */
        service.addComment = function (entryId, content) {
            var comment = new Comment(),
                query = new Parse.Query(Entry),
                user = Parse.User.current();

            return query.get(entryId)
                .then(function (entry) {
                    comment.set('entry', entry);
                    comment.set('createdBy', user);
                    comment.set('content', content);

                    return comment.save()
                        .then( function(comment) {
                            return getCommentData (comment);
                        })
                        .fail(function(error) {
                            return ErrorsService.getErrorMessage(error);
                        });
                })
                .fail(function (error) {
                    return ErrorsService.getErrorMessage(error);
                });
        };

        /**
         * Añade un nuevo paso a la entrada actual.
         * @param entryId {String} Identificador de la entrada a la que se añadirá el paso.
         * @param description {String} Texto con la descripción del paso.
         * @param picData Imagen del paso.
         * @return {Parse.Promise} Promesa que se resolverá cuando el paso se haya guardado.
         */
        service.addStep = function (entryId, description, picData) {
            var step = new Step(),
                query = new Parse.Query(Entry),
                picId = UtilsService.makeId(15),
                base64pic = picData,
                parseFile = new Parse.File(picId + '.jpg', {base64:base64pic});

            return query.get(entryId)
                .then(function (entry) {
                    step.set('entry', entry);
                    step.set('description', description);

                    return parseFile.save()
                        .then(function(parseFile) {
                            // Si la imagen se ha subido correctamente a Parse se guarda la entrada
                            step.set('image', parseFile);
                            return step.save();
                        })
                        .fail(function(error) {
                            // En caso de que haya habido algún error al guardar o leer la imagen de Parse, se devuelve el error
                            return ErrorsService.getErrorMessage(error);
                        });
                })
                .fail(function (error) {
                    return ErrorsService.getErrorMessage(error);
                });
        };


        /**
         * Elimina un paso.
         * @param stepId {String} Identificador del paso a eliminar.
         * @return {Parse.Promise} Promesa que se resolverá cuando el paso se haya eliminado.
         */
        service.deleteStep = function (stepId) {
            var query = new Parse.Query(Step);

            return query.get(stepId)
                .then(function (step) {
                    return step.destroy();
                })
                .fail(function (error) {
                    return ErrorsService.getErrorMessage(error);
                });
        };

        /**
         * Devuelve la entrada actual.
         * @return {Object} Objeto con los datos de la entrada
         */
        service.getCurrentEntry = function () {
            return currentEntry;
        };

        /**
         * Marca o desmarca una entrada como "Me gusta".
         * @param entryId {String} Identificador de la entrada.
         * @return {Number} Número de "likes" que tiene dicha entrada.
         */
        service.likeEntry = function (entryId) {
            return Parse.Cloud.run('entryAction', {entryId: entryId, actionType: 'like'})
                        .then(function (results) {
                            if (results.result === 'ok') {
                                return results.count;
                            }
                        })
                        .fail(function (error) {
                            return ErrorsService.getErrorMessage(error);
                        });
        };

        /**
        * Marca o desmarca una entrada como "Hecho".
        * @param entryId {String} Identificador de la entrada.
        * @return {Number} Número de "Hechos" que tiene dicha entrada.
        */
        service.doneEntry = function (entryId) {
            return Parse.Cloud.run('entryAction', {entryId: entryId, actionType: 'done'})
                        .then(function (results) {
                            if (results.result === 'ok') {
                                return results.count; }
                        })
                        .fail(function (error) {
                            return ErrorsService.getErrorMessage(error);
                        });
        };

        /**
         * Busca entradas.
         * @param text {String} Texto que se desea buscar en las entradas.
         * @param page {Number} Número de página que se desea obtener de la búsqueda.
         * @return {Object} Lista de resultados.
         */
        service.searchEntries = function (text, page) {
            var pageToGet = page || 0;

            return Parse.Cloud.run('getAllEntries', {page: pageToGet, search: text})
                .then(function (results) {
                    return results;
                })
                .fail(function (error) {
                    return ErrorsService.getErrorMessage(error);
                });
        };

        return service;

    }
]);
