'use strict';

/**
* @ngdoc function
* @name rpsApp.service:UserService
* @description
* # UserSerivce
* Servicio para gestionar los usuarios de la aplicación.
*/
rpsApp.factory('UserService', [
    '$rootScope',
    'SessionService',
    'ErrorsService',

    function ($rootScope, SessionService, ErrorsService) {

        var service = {};

        /**
         * Hace login contra los servicios de Parse.
         * @param username {String} Nombre de usuario
         * @param password {String} Contraseña.
         */
        service.login = function (username, password) {
            return Parse.User.logIn(username, password)
                .then(function (user) {
                    // Do stuff after successful login.
                    SessionService.setUser(user);
                })
                .fail(function(error) {
                    // The login failed. Check error to see why.
                    return ErrorsService.getErrorMessage(error);
                });
            };

        /**
         * Hace logout contra los servicios de Parse.
         */
        service.logout = function () {
            Parse.User.logOut();
            SessionService.cleanUser();
        };

        /**
         * Registra un usuario en Parse.com
         * @param registerUser {Object} Objeto con los datos del usuario a registrar.
         */
        service.register = function (registerUser) {
            var user = new Parse.User();
            user.set('email', registerUser.email);
            user.set('username', registerUser.username);
            user.set('password', registerUser.password);

            return user.signUp(null)
                .then(function(user) {
                    user.setACL(new Parse.ACL(Parse.User.current()));
                    user.save();

                    SessionService.setUser(user);
                })
                .fail(function(error) {
                    return ErrorsService.getErrorMessage(error);
                });
        };

        /**
        * Devuelve un objeto con los datos básicos del usuario logueado.
        * @return {Object} Objeto con los datos del usuario.
        */
        service.getUserData = function () {
            console.log(Parse.User.current().attributes);
            return Parse.User.current().attributes;
        };

        /**
         * Devuelve un objeto con los datos básicos del usuario logueado.
         * @return {Object} Objeto con los datos del usuario.
         * @deprecated Unificar este método con getUserData.
         */
        service.getCurrentUser = function () {
            var user = Parse.User.current();

            return { id: user.id, md5Email: user.attributes.md5Email, username: user.attributes.username };
        };

        /**
         * Obtiene los datos básicos de un usuario de Parse según su Id
         * @param userId {String} Identificador del usuario.
         * @return {Object} Objeto con los datos del usuario.
         */
        service.getParseUserById = function (userId) {
            return Parse.Cloud.run('getUserById', {userId: userId})
                .then(function (result) {
                    return result.user;
                })
                .fail(function (error) {
                    return ErrorsService.getErrorMessage(error);
                });
        };

        /**
        * Obtiene los datos básicos de un usuario de Parse según su Id.
        * @param userId {String} Identificador del usuario.
        * @return {Object} Objeto con los datos del usuario.
        * @deprecated Unificar este método con getParseUserById
        */
        service.getUserById = function (userId) {
            var theUser;

            return Parse.Cloud.run('getUserById', {userId: userId})
                .then(function (result) {
                    theUser = { id: result.user.id, md5Email: result.user.attributes.md5Email, username: result.user.attributes.username, following: result.user.attributes.following };

                    return theUser;
                })
                .fail(function (error) {
                    return ErrorsService.getErrorMessage(error);
                });
        };

        /**
         * Hace 'follow' sobre un usuario.
         * @param userId {String} Identificador del usuario a seguir.
         * @return {Parse.Promise} Promesa que se resuelve cuando la acción termina.
         */
        service.followUser = function (userId) {
            var myUser = Parse.User.current(),
                following = myUser.relation('following');

            return service.getParseUserById(userId)
                .then(function (user) {
                    following.add(user);
                    return myUser.save();
                })
                .fail(function (error) {
                    return ErrorsService.getErrorMessage(error);
                });
        };

        /**
        * Hace 'unfollow' sobre un usuario.
        * @param userId {String} Identificador del usuario a dejar de seguir.
        * @return {Parse.Promise} Promesa que se resuelve cuando la acción termina.
        */
        service.unfollowUser = function (userId) {
            var myUser = Parse.User.current(),
                following = myUser.relation('following');

            return service.getParseUserById(userId)
                .then(function (user) {
                    following.remove(user);
                    return myUser.save();
                })
                .fail(function (error) {
                    return ErrorsService.getErrorMessage(error);
                });
        };

        /**
         * Obtiene la lista de mis contactos de Parse.com
         * @return {Object} Lista de contactos.
         */
        service.getMyContacts = function () {

            return Parse.Cloud.run('getUserContacts')
                .then(function (result) {
                    return result.contacts;
                })
                .fail(function (error) {
                    return ErrorsService.getErrorMessage(error);
                });
        };

        /**
         * Obtiene la lista de contactos de un usuario dado.
         * @param userId {String} Identificador del usuario del que queremos consultar sus contactos.
         * @return {Object} Lista de contactos del usuario.
         */
        service.getUserContacts = function (userId) {

            return Parse.Cloud.run('getUserContacts', {userId: userId})
                .then(function (result) {
                    return result.contacts;
                })
                .fail(function (error) {
                    return ErrorsService.getErrorMessage(error);
                });
        };

        /**
        * Obtiene la lista de mis seguidores de Parse.com
        * @return {Object} Lista de seguidores.
        */
        service.getMyFollowers = function () {

            return Parse.Cloud.run('getUserFollowers')
                .then(function (result) {
                    return result.followers;
                })
                .fail(function (error) {
                    return ErrorsService.getErrorMessage(error);
                });
        };

        /**
        * Obtiene la lista de seguidores de un usuario dado.
        * @param userId {String} Identificador del usuario del que queremos consultar sus seguidores.
        * @return {Object} Lista de seguidores del usuario.
        */
        service.getUserFollowers = function (userId) {

            return Parse.Cloud.run('getUserFollowers', {userId: userId})
                .then(function (result) {
                    return result.followers;
                })
                .fail(function (error) {
                    return ErrorsService.getErrorMessage(error);
                });
        };

        /**
         * Busca usuarios en el servidor de Parse.
         * @param text {String} Texto a buscar en los usuarios.
         * @param page {Number} Número de página que queremos obtener.
         * @return {Object} Lista de usuarios obtenida en la búsqueda.
         */
        service.searchUsers = function (text, page) {
            var pageToGet = page || 0;

            return Parse.Cloud.run('searchUsers', {page: pageToGet, search: text})
                .then(function (results) {
                    return results;
                })
                .fail(function (error) {
                    return ErrorsService.getErrorMessage(error);
                });
        };

        return service;

    }
]);
